@@ -32,10 +32,8 @@ cu_closepipe(int argc, void **argv)
 {
 	int *p1 = (int *)argv[0];
 
-	if (p1[0] >= 0)
-		close(p1[0]);
-	if (p1[1] >= 0)
-		close(p1[1]);
+	close(p1[0]);
+	close(p1[1]);
 }
 
 void
@@ -51,8 +49,7 @@ cu_closedir(int argc, void **argv)
 {
 	DIR *d = (DIR *)(argv[0]);
 
-	if (d)
-		closedir(d);
+	closedir(d);
 }
 
 void
@@ -60,7 +57,5 @@ cu_closefd(int argc, void **argv)
 {
 	int ip = *(int *)argv[0];
 
-	if (ip >= 0)
-		close(ip);
+	close(ip);
 }
-
