@@ -3,7 +3,9 @@
  * main.c - main program
  *
  * Copyright © 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
- * Copyright © 2006-2009 Guillem Jover <guillem@debian.org>
+ * Copyright © 2006-2010 Guillem Jover <guillem@debian.org>
+ * Copyright © 2010 Canonical Ltd.
+ *   written by Martin Pitt <martin.pitt@canonical.com>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -35,7 +37,6 @@
 #include <string.h>
 #include <fcntl.h>
 #include <dirent.h>
-#include <signal.h>
 #include <unistd.h>
 #include <stdbool.h>
 #include <stdlib.h>
@@ -45,30 +46,37 @@
 #include <dpkg/i18n.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
+#include <dpkg/arch.h>
+#include <dpkg/subproc.h>
+#include <dpkg/command.h>
 #include <dpkg/myopt.h>
+#include <dpkg/pkg-spec.h>
 
 #include "main.h"
 #include "filesdb.h"
+#include "filters.h"
 
-static void
+static void DPKG_ATTR_NORET
 printversion(const struct cmdinfo *ci, const char *value)
 {
   printf(_("Debian `%s' package management program version %s.\n"),
          DPKG, DPKG_VERSION_ARCH);
   printf(_(
 "This is free software; see the GNU General Public License version 2 or\n"
-"later for copying conditions. There is NO warranty.\n"
-"See %s --license for copyright and license details.\n"), DPKG);
+"later for copying conditions. There is NO warranty.\n"));
 
   m_output(stdout, _("<standard output>"));
 
   exit(0);
 }
+
 /*
-   options that need fixing:
-  dpkg --yet-to-unpack                 \n\
-  */
-static void
+ * FIXME: Options that need fixing:
+ * dpkg --yet-to-unpack
+ * dpkg --command-fd
+ */
+
+static void DPKG_ATTR_NORET
 usage(const struct cmdinfo *ci, const char *value)
 {
   printf(_(
@@ -106,7 +114,6 @@ usage(const struct cmdinfo *ci, const char *value)
   printf(_(
 "  -h|--help                        Show this help message.\n"
 "  --version                        Show the version.\n"
-"  --license|--licence              Show the copyright licensing terms.\n"
 "\n"));
 
   printf(_(
@@ -116,7 +123,8 @@ usage(const struct cmdinfo *ci, const char *value)
 
   printf(_(
 "For internal use: dpkg --assert-support-predepends | --predep-package |\n"
-"  --assert-working-epoch | --assert-long-filenames | --assert-multi-conrep.\n"
+"  --assert-working-epoch | --assert-long-filenames | --assert-multi-conrep |\n"
+"  --assert-multi-arch.\n"
 "\n"));
 
   printf(_(
@@ -124,6 +132,8 @@ usage(const struct cmdinfo *ci, const char *value)
 "  --admindir=<directory>     Use <directory> instead of %s.\n"
 "  --root=<directory>         Install on a different root directory.\n"
 "  --instdir=<directory>      Change installation dir without changing admin dir.\n"
+"  --path-exclude=<pattern>   Do not install paths which match a shell pattern.\n"
+"  --path-include=<pattern>   Re-include a pattern after a previous exclusion.\n"
 "  -O|--selected-only         Skip packages not selected for install/upgrade.\n"
 "  -E|--skip-same-version     Skip packages whose same version is installed.\n"
 "  -G|--refuse-downgrade      Skip packages with earlier version than installed.\n"
@@ -159,117 +169,176 @@ usage(const struct cmdinfo *ci, const char *value)
 }
 
 const char thisname[]= "dpkg";
-const char architecture[]= ARCHITECTURE;
 const char printforhelp[]= N_(
 "Type dpkg --help for help about installing and deinstalling packages [*];\n"
 "Use `dselect' or `aptitude' for user-friendly package management;\n"
 "Type dpkg -Dhelp for a list of dpkg debug flag values;\n"
 "Type dpkg --force-help for a list of forcing options;\n"
 "Type dpkg-deb --help for help about manipulating *.deb files;\n"
-"Type dpkg --license for copyright license and lack of warranty (GNU GPL) [*].\n"
 "\n"
 "Options marked [*] produce a lot of output - pipe it through `less' or `more' !");
 
-const struct cmdinfo *cipaction = NULL;
 int f_pending=0, f_recursive=0, f_alsoselect=1, f_skipsame=0, f_noact=0;
 int f_autodeconf=0, f_nodebsig=0;
 int f_triggers = 0;
-unsigned long f_debug=0;
-/* Change fc_overwrite to 1 to enable force-overwrite by default */
 int fc_downgrade=1, fc_configureany=0, fc_hold=0, fc_removereinstreq=0, fc_overwrite=0;
 int fc_removeessential=0, fc_conflicts=0, fc_depends=0, fc_dependsversion=0;
 int fc_breaks=0, fc_badpath=0, fc_overwritediverted=0, fc_architecture=0;
 int fc_nonroot=0, fc_overwritedir=0, fc_conff_new=0, fc_conff_miss=0;
 int fc_conff_old=0, fc_conff_def=0;
+int fc_conff_ask = 0;
+int fc_unsafe_io = 0;
 int fc_badverify = 0;
 
 int errabort = 50;
-const char *admindir= ADMINDIR;
+static const char *admindir = ADMINDIR;
 const char *instdir= "";
 struct pkg_list *ignoredependss = NULL;
 
+static const char *
+forcetype_str(char type)
+{
+  switch (type) {
+  case '\0':
+  case ' ':
+    return "   ";
+  case '*':
+    return "[*]";
+  case '!':
+    return "[!]";
+  default:
+    internerr("unknown force type %c", type);
+  }
+}
+
 static const struct forceinfo {
   const char *name;
   int *opt;
+  char type;
+  const char *desc;
 } forceinfos[]= {
-  { "downgrade",           &fc_downgrade                },
-  { "configure-any",       &fc_configureany             },
-  { "hold",                &fc_hold                     },
-  { "remove-reinstreq",    &fc_removereinstreq          },
-  { "remove-essential",    &fc_removeessential          },
-  { "conflicts",           &fc_conflicts                },
-  { "confnew",             &fc_conff_new                },
-  { "confold",             &fc_conff_old                },
-  { "confdef",             &fc_conff_def                },
-  { "confmiss",            &fc_conff_miss               },
-  { "depends",             &fc_depends                  },
-  { "depends-version",     &fc_dependsversion           },
-  { "breaks",              &fc_breaks                   },
-  { "bad-path",            &fc_badpath                  },
-  { "not-root",            &fc_nonroot                  },
-  { "overwrite",           &fc_overwrite                },
-  { "overwrite-diverted",  &fc_overwritediverted        },
-  { "overwrite-dir",       &fc_overwritedir             },
-  { "architecture",        &fc_architecture             },
-  { "bad-verify",          &fc_badverify                },
-  {  NULL                                               }
+  { "all",                 NULL,
+    '!', N_("Set all force options")},
+  { "downgrade",           &fc_downgrade,
+    '*', N_("Replace a package with a lower version") },
+  { "configure-any",       &fc_configureany,
+    ' ', N_("Configure any package which may help this one") },
+  { "hold",                &fc_hold,
+    ' ', N_("Process incidental packages even when on hold") },
+  { "not-root",            &fc_nonroot,
+    ' ', N_("Try to (de)install things even when not root") },
+  { "bad-path",            &fc_badpath,
+    ' ', N_("PATH is missing important programs, problems likely") },
+  { "bad-verify",          &fc_badverify,
+    ' ', N_("Install a package even if it fails authenticity check") },
+  { "overwrite",           &fc_overwrite,
+    ' ', N_("Overwrite a file from one package with another") },
+  { "overwrite-diverted",  &fc_overwritediverted,
+    ' ', N_("Overwrite a diverted file with an undiverted version") },
+  { "overwrite-dir",       &fc_overwritedir,
+    '!', N_("Overwrite one package's directory with another's file") },
+  { "unsafe-io",           &fc_unsafe_io,
+    '!', N_("Do not perform safe I/O operations when unpacking") },
+  { "confnew",             &fc_conff_new,
+    '!', N_("Always use the new config files, don't prompt") },
+  { "confold",             &fc_conff_old,
+    '!', N_("Always use the old config files, don't prompt") },
+  /* XXX: Handle automatically line wrappings. */
+  { "confdef",             &fc_conff_def,
+    '!', N_("Use the default option for new config files if one\n"
+"                         is available, don't prompt. If no default can be found,\n"
+"                         you will be prompted unless one of the confold or\n"
+"                         confnew options is also given") },
+  { "confmiss",            &fc_conff_miss,
+    '!', N_("Always install missing config files") },
+  { "confask",             &fc_conff_ask,
+    '!', N_("Offer to replace config files with no new versions") },
+  { "architecture",        &fc_architecture,
+    '!', N_("Process even packages with wrong or no architecture") },
+  { "breaks",              &fc_breaks,
+    '!', N_("Install even if it would break another package") },
+  { "conflicts",           &fc_conflicts,
+    '!', N_("Allow installation of conflicting packages") },
+  { "depends",             &fc_depends,
+    '!', N_("Turn all dependency problems into warnings") },
+  { "depends-version",     &fc_dependsversion,
+    '!', N_("Turn dependency version problems into warnings") },
+  { "remove-reinstreq",    &fc_removereinstreq,
+    '!', N_("Remove packages which require installation") },
+  { "remove-essential",    &fc_removeessential,
+    '!', N_("Remove an essential package") },
+  { NULL }
 };
 
-static void setaction(const struct cmdinfo *cip, const char *value) {
-  if (cipaction)
-    badusage(_("conflicting actions -%c (--%s) and -%c (--%s)"),
-             cip->oshort, cip->olong, cipaction->oshort, cipaction->olong);
-  cipaction= cip;
-}
+#define DBG_DEF(n, d) \
+  { .flag = dbg_##n, .name = #n, .desc = d }
 
-static void setobsolete(const struct cmdinfo *cip, const char *value) {
-  warning(_("obsolete option '--%s'\n"), cip->olong);
-}
+static const struct debuginfo {
+  int flag;
+  const char *name;
+  const char *desc;
+} debuginfos[] = {
+  DBG_DEF(general,         N_("Generally helpful progress information")),
+  DBG_DEF(scripts,         N_("Invocation and status of maintainer scripts")),
+  DBG_DEF(eachfile,        N_("Output for each file processed")),
+  DBG_DEF(eachfiledetail,  N_("Lots of output for each file processed")),
+  DBG_DEF(conff,           N_("Output for each configuration file")),
+  DBG_DEF(conffdetail,     N_("Lots of output for each configuration file")),
+  DBG_DEF(depcon,          N_("Dependencies and conflicts")),
+  DBG_DEF(depcondetail,    N_("Lots of dependencies/conflicts output")),
+  DBG_DEF(triggers,        N_("Trigger activation and processing")),
+  DBG_DEF(triggersdetail,  N_("Lots of output regarding triggers")),
+  DBG_DEF(triggersstupid,  N_("Silly amounts of output regarding triggers")),
+  DBG_DEF(veryverbose,     N_("Lots of drivel about eg the dpkg/info directory")),
+  DBG_DEF(stupidlyverbose, N_("Insane amounts of drivel")),
+  { 0, NULL, NULL }
+};
 
 static void setdebug(const struct cmdinfo *cpi, const char *value) {
   char *endp;
+  unsigned long mask;
+  const struct debuginfo *dip;
 
   if (*value == 'h') {
     printf(_(
 "%s debugging option, --debug=<octal> or -D<octal>:\n"
 "\n"
-" number  ref. in source   description\n"
-"      1   general           Generally helpful progress information\n"
-"      2   scripts           Invocation and status of maintainer scripts\n"
-"     10   eachfile          Output for each file processed\n"
-"    100   eachfiledetail    Lots of output for each file processed\n"
-"     20   conff             Output for each configuration file\n"
-"    200   conffdetail       Lots of output for each configuration file\n"
-"     40   depcon            Dependencies and conflicts\n"
-"    400   depcondetail      Lots of dependencies/conflicts output\n"
-"  10000   triggers          Trigger activation and processing\n"
-"  20000   triggersdetail    Lots of output regarding triggers\n"
-"  40000   triggersstupid    Silly amounts of output regarding triggers\n"
-"   1000   veryverbose       Lots of drivel about eg the dpkg/info directory\n"
-"   2000   stupidlyverbose   Insane amounts of drivel\n"
-"\n"
-"Debugging options are be mixed using bitwise-or.\n"
-"Note that the meanings and values are subject to change.\n"), DPKG);
+" Number  Ref. in source   Description\n"), DPKG);
+
+    for (dip = debuginfos; dip->name; dip++)
+      printf(" %6o  %-16s %s\n", dip->flag, dip->name, gettext(dip->desc));
+
+    printf(_("\n"
+"Debugging options can be mixed using bitwise-or.\n"
+"Note that the meanings and values are subject to change.\n"));
     m_output(stdout, _("<standard output>"));
     exit(0);
   }
-  
-  f_debug= strtoul(value,&endp,8);
+
+  mask = strtoul(value, &endp, 8);
   if (value == endp || *endp) badusage(_("--debug requires an octal argument"));
+
+  debug_set_mask(mask);
+}
+
+static void
+setfilter(const struct cmdinfo *cip, const char *value)
+{
+  filter_add(value, cip->arg_int);
 }
 
 static void setroot(const struct cmdinfo *cip, const char *value) {
   char *p;
   instdir= value;
-  p= m_malloc(strlen(value) + sizeof(ADMINDIR));
-  strcpy(p,value);
-  strcat(p,ADMINDIR);
+  m_asprintf(&p, "%s%s", value, ADMINDIR);
   admindir= p;
 }
 
 static void ignoredepends(const struct cmdinfo *cip, const char *value) {
   char *copy, *p;
   const char *pnerr;
+  struct pkginfo *pkg;
+  struct pkg_spec ps = PKG_SPEC_INIT(psf_def_native | psf_no_check | psf_patterns);
 
   copy= m_malloc(strlen(value)+2);
   strcpy(copy,value);
@@ -283,14 +352,18 @@ static void ignoredepends(const struct cmdinfo *cip, const char *value) {
   }
   p= copy;
   while (*p) {
-    pnerr = illegal_packagename(p, NULL);
+    pkg_spec_parse(&ps, p);
+    pnerr = pkg_spec_is_illegal(&ps);
     if (pnerr) ohshite(_("--ignore-depends requires a legal package name. "
                        "`%.250s' is not; %s"), p, pnerr);
 
-    pkg_list_prepend(&ignoredependss, findpackage(p));
+    pkg_spec_iter_start(&ps);
+    while ((pkg = pkg_spec_iter_next_pkg(&ps)))
+      pkg_list_prepend(&ignoredependss, pkg);
 
     p+= strlen(p)+1;
   }
+  pkg_spec_reset(&ps);
 
   free(copy);
 }
@@ -306,8 +379,6 @@ static void setinteger(const struct cmdinfo *cip, const char *value) {
 }
 
 static void setpipe(const struct cmdinfo *cip, const char *value) {
-  struct pipef **pipe_head = cip->parg;
-  struct pipef *pipe_new;
   unsigned long v;
   char *ep;
 
@@ -315,12 +386,7 @@ static void setpipe(const struct cmdinfo *cip, const char *value) {
   if (value == ep || *ep || v > INT_MAX)
     badusage(_("invalid integer for --%s: `%.250s'"),cip->olong,value);
 
-  setcloexec(v, _("<package status and progress file descriptor>"));
-
-  pipe_new = nfmalloc(sizeof(struct pipef));
-  pipe_new->fd = v;
-  pipe_new->next = *pipe_head;
-  *pipe_head = pipe_new;
+  statusfd_add(v);
 }
 
 static bool
@@ -343,11 +409,13 @@ struct invoke_hook *pre_invoke_hooks = NULL;
 struct invoke_hook **pre_invoke_hooks_tail = &pre_invoke_hooks;
 struct invoke_hook *post_invoke_hooks = NULL;
 struct invoke_hook **post_invoke_hooks_tail = &post_invoke_hooks;
+struct invoke_hook *status_loggers = NULL;
+struct invoke_hook **status_loggers_tail = &status_loggers;
 
 static void
 set_invoke_hook(const struct cmdinfo *cip, const char *value)
 {
-  struct invoke_hook ***hook_tail = cip->parg;
+  struct invoke_hook ***hook_tail = cip->arg_ptr;
   struct invoke_hook *hook_new;
 
   hook_new = nfmalloc(sizeof(struct invoke_hook));
@@ -373,12 +441,57 @@ run_invoke_hooks(const char *action, struct invoke_hook *hook_head)
      * used “!$=&|\\`'"^~;<>{}[]()?*#”. */
     status = system(hook->command);
     if (status != 0)
-      ohshit("error executing hook '%s', exit code %d", hook->command, status);
+      ohshit(_("error executing hook '%s', exit code %d"), hook->command,
+             status);
   }
 
   unsetenv("DPKG_HOOK_ACTION");
 }
 
+static void
+run_status_loggers(struct invoke_hook *hook_head)
+{
+  struct invoke_hook *hook;
+
+  for (hook = hook_head; hook; hook = hook->next) {
+    pid_t pid;
+    int p[2];
+
+    m_pipe(p);
+
+    pid = subproc_fork();
+    if (pid == 0) {
+      /* Setup stdin and stdout. */
+      m_dup2(p[0], 0);
+      close(1);
+
+      close(p[0]);
+      close(p[1]);
+
+      command_shell(hook->command, _("status logger"));
+    }
+    close(p[0]);
+
+    statusfd_add(p[1]);
+  }
+}
+
+static void
+set_foreign_arch(const struct cmdinfo *cip, const char *value)
+{
+  struct dpkg_arch *arch;
+
+  arch = dpkg_arch_find(value);
+  if (arch->type == arch_unknown || arch->type == arch_foreign)
+    arch->type = arch_foreign;
+  else if (arch->type == arch_illegal)
+    warning(_("ignoring option --%s=%s: %s"), cip->olong, value,
+            dpkg_arch_name_is_illegal(value));
+  else
+    warning(_("ignoring option --%s=%s: %s"), cip->olong, value,
+            _("this architecture cannot be foreign"));
+}
+
 static void setforce(const struct cmdinfo *cip, const char *value) {
   const char *comma;
   size_t l;
@@ -389,34 +502,16 @@ static void setforce(const struct cmdinfo *cip, const char *value) {
 "%s forcing options - control behaviour when problems found:\n"
 "  warn but continue:  --force-<thing>,<thing>,...\n"
 "  stop with error:    --refuse-<thing>,<thing>,... | --no-force-<thing>,...\n"
-" Forcing things:\n"
-"  all [!]                Set all force options\n"
-"  downgrade [*]          Replace a package with a lower version\n"
-"  configure-any          Configure any package which may help this one\n"
-"  hold                   Process incidental packages even when on hold\n"
-"  bad-path               PATH is missing important programs, problems likely\n"
-"  not-root               Try to (de)install things even when not root\n"
-"  overwrite              Overwrite a file from one package with another\n"
-"  overwrite-diverted     Overwrite a diverted file with an undiverted version\n"
-"  bad-verify             Install a package even if it fails authenticity check\n"
-"  depends-version [!]    Turn dependency version problems into warnings\n"
-"  depends [!]            Turn all dependency problems into warnings\n"
-"  confnew [!]            Always use the new config files, don't prompt\n"
-"  confold [!]            Always use the old config files, don't prompt\n"
-"  confdef [!]            Use the default option for new config files if one\n"
-"                         is available, don't prompt. If no default can be found,\n"
-"                         you will be prompted unless one of the confold or\n"
-"                         confnew options is also given\n"
-"  confmiss [!]           Always install missing config files\n"
-"  breaks [!]             Install even if it would break another package\n"
-"  conflicts [!]          Allow installation of conflicting packages\n"
-"  architecture [!]       Process even packages with wrong architecture\n"
-"  overwrite-dir [!]      Overwrite one package's directory with another's file\n"
-"  remove-reinstreq [!]   Remove packages which require installation\n"
-"  remove-essential [!]   Remove an essential package\n"
+" Forcing things:\n"), DPKG);
+
+    for (fip = forceinfos; fip->name; fip++)
+      printf("  %s %-18s %s\n", forcetype_str(fip->type), fip->name,
+             gettext(fip->desc));
+
+    printf(_(
 "\n"
 "WARNING - use of options marked [!] can seriously damage your installation.\n"
-"Forcing options marked [*] are enabled by default.\n"), DPKG);
+"Forcing options marked [*] are enabled by default.\n"));
     m_output(stdout, _("<standard output>"));
     exit(0);
   }
@@ -426,38 +521,34 @@ static void setforce(const struct cmdinfo *cip, const char *value) {
     l = comma ? (size_t)(comma - value) : strlen(value);
     for (fip=forceinfos; fip->name; fip++)
       if (!strncmp(fip->name,value,l) && strlen(fip->name)==l) break;
+
     if (!fip->name) {
-      if (strncmp("all", value, l) == 0) {
-	for (fip=forceinfos; fip->name; fip++)
-	  if (fip->opt)
-	    *fip->opt= cip->arg;
-      } else
-	badusage(_("unknown force/refuse option `%.*s'"),
-	         (int)min(l, 250), value);
+      badusage(_("unknown force/refuse option `%.*s'"),
+               (int)min(l, 250), value);
+    } else if (strcmp(fip->name, "all") == 0) {
+      for (fip = forceinfos; fip->name; fip++)
+        if (fip->opt)
+          *fip->opt = cip->arg_int;
+    } else if (fip->opt) {
+      *fip->opt = cip->arg_int;
     } else {
-      if (fip->opt)
-	*fip->opt= cip->arg;
-      else
-	warning(_("obsolete force/refuse option '%s'\n"), fip->name);
+      warning(_("obsolete force/refuse option '%s'\n"), fip->name);
     }
+
     if (!comma) break;
     value= ++comma;
   }
 }
 
-void execbackend(const char *const *argv) DPKG_ATTR_NORET;
-void commandfd(const char *const *argv);
+int execbackend(const char *const *argv) DPKG_ATTR_NORET;
+int commandfd(const char *const *argv);
+
+/* This table has both the action entries in it and the normal options.
+ * The action entries are made with the ACTION macro, as they all
+ * have a very similar structure. */
 static const struct cmdinfo cmdinfos[]= {
-  /* This table has both the action entries in it and the normal options.
-   * The action entries are made with the ACTION macro, as they all
-   * have a very similar structure.
-   */
-#define ACTION(longopt,shortopt,code,function) \
- { longopt, shortopt, 0, NULL, NULL, setaction, code, NULL, (voidfnp)function }
-#define OBSOLETE(longopt,shortopt) \
- { longopt, shortopt, 0, NULL, NULL, setobsolete, 0, NULL, NULL }
 #define ACTIONBACKEND(longopt, shortopt, backend) \
- { longopt, shortopt, 0, NULL, NULL, setaction, 0, (void *)backend, (voidfnp)execbackend }
+ { longopt, shortopt, 0, NULL, NULL, setaction, 0, (void *)backend, execbackend }
 
   ACTION( "install",                        'i', act_install,              archivefiles    ),
   ACTION( "unpack",                          0,  act_unpack,               archivefiles    ),
@@ -484,17 +575,23 @@ static const struct cmdinfo cmdinfos[]= {
   ACTION( "assert-working-epoch",            0,  act_assertepoch,          assertepoch     ),
   ACTION( "assert-long-filenames",           0,  act_assertlongfilenames,  assertlongfilenames ),
   ACTION( "assert-multi-conrep",             0,  act_assertmulticonrep,    assertmulticonrep ),
+  ACTION( "assert-multi-arch",               0,  act_assertmultiarch,      assertmultiarch ),
   ACTION( "print-architecture",              0,  act_printarch,            printarch   ),
   ACTION( "print-installation-architecture", 0,  act_printinstarch,        printinstarch  ),
+  ACTION( "print-foreign-architectures",     0,  act_foreignarchs,         print_foreign_archs ),
   ACTION( "predep-package",                  0,  act_predeppackage,        predeppackage   ),
   ACTION( "compare-versions",                0,  act_cmpversions,          cmpversions     ),
 /*
   ACTION( "command-fd",                   'c', act_commandfd,   commandfd     ),
 */
-  
+
   { "pre-invoke",        0,   1, NULL,          NULL,      set_invoke_hook, 0, &pre_invoke_hooks_tail },
   { "post-invoke",       0,   1, NULL,          NULL,      set_invoke_hook, 0, &post_invoke_hooks_tail },
-  { "status-fd",         0,   1, NULL,          NULL,      setpipe, 0, &status_pipes },
+  { "path-exclude",      0,   1, NULL,          NULL,      setfilter,     0 },
+  { "path-include",      0,   1, NULL,          NULL,      setfilter,     1 },
+  { "foreign-architecture", 0, 1, NULL,         NULL,      set_foreign_arch, 0 },
+  { "status-logger",     0,   1, NULL,          NULL,      set_invoke_hook, 0, &status_loggers_tail },
+  { "status-fd",         0,   1, NULL,          NULL,      setpipe, 0 },
   { "log",               0,   1, NULL,          &log_file, NULL,    0 },
   { "pending",           'a', 0, &f_pending,    NULL,      NULL,    1 },
   { "recursive",         'R', 0, &f_recursive,  NULL,      NULL,    1 },
@@ -522,10 +619,6 @@ static const struct cmdinfo cmdinfos[]= {
   { "debug",             'D', 1, NULL,          NULL,      setdebug,      0 },
   { "help",              'h', 0, NULL,          NULL,      usage,         0 },
   { "version",           0,   0, NULL,          NULL,      printversion,  0 },
-  /* UK spelling. */
-  { "licence",           0,   0, NULL,          NULL,      showcopyright, 0 },
-  /* US spelling. */
-  { "license",           0,   0, NULL,          NULL,      showcopyright, 0 },
   ACTIONBACKEND( "build",		'b', BACKEND),
   ACTIONBACKEND( "contents",		'c', BACKEND),
   ACTIONBACKEND( "control",		'e', BACKEND),
@@ -539,69 +632,38 @@ static const struct cmdinfo cmdinfos[]= {
   { NULL,                0,   0, NULL,          NULL,      NULL,          0 }
 };
 
-void execbackend(const char *const *argv) {
-  char **nargv;   /* argv for backend command */
-  int i = 0;      /* index for nargv */
-  int offset = 0; /* offset for copying argv strings to nargv */
-  int argc = 1;   /* for nargv */
-  const char *const *arg = argv;
-  int pass_admindir = 0;
+int
+execbackend(const char *const *argv)
+{
+  struct command cmd;
+  char *arg;
 
-  while (*arg != NULL) {
-    arg++; argc++;
-  }
+  command_init(&cmd, cipaction->arg_ptr, NULL);
+  command_add_arg(&cmd, cipaction->arg_ptr);
 
-  /*
-   * Special case: dpkg-query takes the --admindir option, and if dpkg itself
-   * was given a different admin directory, we need to pass it along to it.
-   */
-  if (strcmp(cipaction->parg, DPKGQUERY) == 0 &&
-      strcmp(admindir, ADMINDIR) != 0) {
-    argc++;
-    pass_admindir = 1;
-  }
-
-  nargv = m_malloc(sizeof(char *) * (argc + 3));
-  nargv[i] = m_strdup(cipaction->parg);
-
-  i++, offset++;
-
-  if (pass_admindir) {
-    nargv[i] = m_malloc((strlen("--admindir=") + strlen(admindir) + 1));
-    sprintf(nargv[i], "--admindir=%s", admindir);
-    i++, offset++;
-  }
-
-  nargv[i] = m_malloc(2 + strlen(cipaction->olong) + 1);
-  strcpy(nargv[i], "--");
-  strcat(nargv[i], cipaction->olong);
-  i++, offset++;
+  m_asprintf(&arg, "--%s", cipaction->olong);
+  command_add_arg(&cmd, arg);
 
   /* Exlicitely separate arguments from options as any user-supplied
    * separator got stripped by the option parser */
-  nargv[i] = "--";
-  i++, argc++, offset++;
+  command_add_arg(&cmd, "--");
+  command_add_argl(&cmd, (const char **)argv);
 
-  /* Copy arguments from argv to nargv. */
-  for (; i <= argc; i++)
-    nargv[i] = m_strdup(argv[i - offset]);
-
-  nargv[i] = NULL;
-
-  execvp(cipaction->parg, nargv);
-  ohshite(_("failed to exec %s"), (char *)cipaction->parg);
+  command_exec(&cmd);
 }
 
-void commandfd(const char *const *argv) {
-  jmp_buf ejbuf;
+int
+commandfd(const char *const *argv)
+{
   struct varbuf linevb = VARBUF_INIT;
   const char * pipein;
-  const char **newargs;
+  const char **newargs = NULL;
   char *ptr, *endptr;
   FILE *in;
   unsigned long infd;
-  int c, lno, i, skipchar;
-  static void (*actionfunction)(const char *const *argv);
+  int ret = 0;
+  int c, lno, i;
+  bool skipchar;
 
   pipein = *argv++;
   if (pipein == NULL)
@@ -615,15 +677,12 @@ void commandfd(const char *const *argv) {
   if ((in= fdopen(infd, "r")) == NULL)
     ohshite(_("couldn't open `%i' for stream"), (int) infd);
 
-  if (setjmp(ejbuf)) { /* expect warning about possible clobbering of argv */
-    error_unwind(ehflag_bombout); exit(2);
-  }
-
-  for (;;lno= 0) {
-    const char **oldargs= NULL;
-    int argc= 1, mode= 0;
+  for (;;) {
+    bool mode = false;
+    int argc= 1;
     lno= 0;
-    push_error_handler(&ejbuf, print_error_fatal, NULL);
+
+    push_error_context();
 
     do { c= getc(in); if (c == '\n') lno++; } while (c != EOF && isspace(c));
     if (c == EOF) break;
@@ -631,39 +690,42 @@ void commandfd(const char *const *argv) {
       do { c= getc(in); if (c == '\n') lno++; } while (c != EOF && c != '\n');
       continue;
     }
-    varbufreset(&linevb);
+    varbuf_reset(&linevb);
     do {
-      varbufaddc(&linevb,c);
+      varbuf_add_char(&linevb, c);
       c= getc(in);
       if (c == '\n') lno++;
-      if (isspace(c)) argc++;  /* This isn't fully accurate, but overestimating can't hurt. */
+
+      /* This isn't fully accurate, but overestimating can't hurt. */
+      if (isspace(c))
+        argc++;
     } while (c != EOF && c != '\n');
     if (c == EOF) ohshit(_("unexpected eof before end of line %d"),lno);
     if (!argc) continue;
-    varbufaddc(&linevb,0);
-    oldargs = newargs = m_realloc(oldargs, sizeof(const char *) * (argc + 1));
+    varbuf_end_str(&linevb);
+    newargs = m_realloc(newargs, sizeof(const char *) * (argc + 1));
     argc= 1;
     ptr= linevb.buf;
-    endptr= ptr + linevb.used;
-    skipchar= 0;
+    endptr = ptr + linevb.used + 1;
+    skipchar = false;
     while(ptr < endptr) {
       if (skipchar) {
-	skipchar= 0;
+	skipchar = false;
       } else if (*ptr == '\\') {
 	memmove(ptr, (ptr+1), (linevb.used-(linevb.buf - ptr)-1));
 	endptr--;
-	skipchar= 1;
+	skipchar = true;
 	continue;
       } else if (isspace(*ptr)) {
-	if (mode == 1) {
+	if (mode == true) {
 	  *ptr = '\0';
-	  mode= 0;
+	  mode = false;
 	}
       } else {
-	if (mode == 0) {
+	if (mode == false) {
 	  newargs[argc]= ptr;
 	  argc++;
-	  mode= 1;
+	  mode = true;
 	}
       }
       ptr++;
@@ -671,59 +733,64 @@ void commandfd(const char *const *argv) {
     *ptr = '\0';
     newargs[argc++] = NULL;
 
-/* We strdup each argument, but never free it, because the error messages
- * contain references back to these strings.  Freeing them, and reusing
- * the memory, would make those error messages confusing, to say the
- * least.
- */
+    /* We strdup() each argument, but never free it, because the
+     * error messages contain references back to these strings.
+     * Freeing them, and reusing the memory, would make those
+     * error messages confusing, to say the least. */
     for(i=1;i<argc;i++)
       if (newargs[i])
         newargs[i] = m_strdup(newargs[i]);
 
-    cipaction= NULL;
+    setaction(NULL, NULL);
     myopt((const char *const**)&newargs,cmdinfos);
     if (!cipaction) badusage(_("need an action option"));
 
-    actionfunction= (void (*)(const char* const*))cipaction->farg;
-    actionfunction(newargs);
-    set_error_display(NULL, NULL);
-    error_unwind(ehflag_normaltidy);
+    ret |= cipaction->action(newargs);
+
+    pop_error_context(ehflag_normaltidy);
   }
+
+  return ret;
 }
 
-
 int main(int argc, const char *const *argv) {
-  jmp_buf ejbuf;
-  static void (*actionfunction)(const char *const *argv);
+  int ret;
 
-  setlocale(LC_ALL, "");
+  if (getenv("DPKG_UNTRANSLATED_MESSAGES") == NULL)
+     setlocale(LC_ALL, "");
   bindtextdomain(PACKAGE, LOCALEDIR);
   textdomain(PACKAGE);
 
-  standard_startup(&ejbuf);
+  standard_startup();
   loadcfgfile(DPKG, cmdinfos);
   myopt(&argv, cmdinfos);
 
   if (!cipaction) badusage(_("need an action option"));
 
+  admindir = dpkg_db_set_dir(admindir);
+
+  /* Always set environment, to avoid possible security risks. */
+  if (setenv("DPKG_ADMINDIR", admindir, 1) < 0)
+    ohshite(_("unable to setenv for subprocesses"));
+
   if (!f_triggers)
-    f_triggers = (cipaction->arg == act_triggers && *argv) ? -1 : 1;
+    f_triggers = (cipaction->arg_int == act_triggers && *argv) ? -1 : 1;
 
   setvbuf(stdout, NULL, _IONBF, 0);
 
-  if (is_invoke_action(cipaction->arg))
+  if (is_invoke_action(cipaction->arg_int)) {
     run_invoke_hooks(cipaction->olong, pre_invoke_hooks);
+    run_status_loggers(status_loggers);
+  }
 
   filesdbinit();
 
-  actionfunction= (void (*)(const char* const*))cipaction->farg;
+  ret = cipaction->action(argv);
 
-  actionfunction(argv);
-
-  if (is_invoke_action(cipaction->arg))
+  if (is_invoke_action(cipaction->arg_int))
     run_invoke_hooks(cipaction->olong, post_invoke_hooks);
 
   standard_shutdown();
 
-  return reportbroken_retexitstatus();
+  return reportbroken_retexitstatus(ret);
 }
