@@ -131,7 +131,7 @@ enqueue_specified(const char *const *argv)
   pkg_spec_reset(&pkgspec);
 }
 
-void
+int
 packages(const char *const *argv)
 {
   enum modstatdb_rw msdb_status;
@@ -163,6 +163,8 @@ packages(const char *const *argv)
   trigproc_run_deferred();
 
   modstatdb_shutdown();
+
+  return 0;
 }
 
 void process_queue(void) {
