@@ -21,6 +21,7 @@
  */
 
 #include <config.h>
+#include <compat.h>
 
 #include <dpkg/macros.h>
 
@@ -115,7 +116,7 @@
 #endif
 
 #if defined(OSLinux)
-/* This comes from TASK_COMM_LEN defined in linux's include/linux/sched.h. */
+/* This comes from TASK_COMM_LEN defined in Linux's include/linux/sched.h. */
 #define PROCESS_NAME_SIZE 15
 #endif
 
@@ -192,9 +193,11 @@ struct schedule_item {
 		sched_timeout,
 		sched_signal,
 		sched_goto,
-		sched_forever /* Only seen within parse_schedule and callees */
+		/* Only seen within parse_schedule and callees. */
+		sched_forever,
 	} type;
-	int value; /* Seconds, signal no., or index into array. */
+	/* Seconds, signal no., or index into array. */
+	int value;
 };
 
 static struct res_schedule *proc_sched = NULL;
@@ -203,27 +206,8 @@ static struct res_schedule *io_sched = NULL;
 static int schedule_length;
 static struct schedule_item *schedule = NULL;
 
-static void *xmalloc(int size);
-static void push(struct pid_list **list, pid_t pid);
-static void do_help(void);
-static void parse_options(int argc, char * const *argv);
-static bool pid_is_user(pid_t pid, uid_t uid);
-static bool pid_is_cmd(pid_t pid, const char *name);
-static void check(pid_t pid);
-static void do_pidfile(const char *name);
-static void do_stop(int signal_nr, int quietmode,
-                    int *n_killed, int *n_notkilled, int retry_nr);
-#if defined(OSLinux) || defined(OShpux)
-static bool pid_is_exec(pid_t pid, const struct stat *esb);
-#endif
 
-
-static void fatal(const char *format, ...)
-	DPKG_ATTR_NORET DPKG_ATTR_PRINTF(1);
-static void badusage(const char *msg)
-	DPKG_ATTR_NORET;
-
-static void
+static void DPKG_ATTR_PRINTF(1)
 warning(const char *format, ...)
 {
 	va_list arglist;
@@ -234,7 +218,7 @@ warning(const char *format, ...)
 	va_end(arglist);
 }
 
-static void
+static void DPKG_ATTR_NORET DPKG_ATTR_PRINTF(1)
 fatal(const char *format, ...)
 {
 	va_list arglist;
@@ -278,7 +262,7 @@ static void
 xgettimeofday(struct timeval *tv)
 {
 	if (gettimeofday(tv, NULL) != 0)
-		fatal("gettimeofday failed: %s", strerror(errno));
+		fatal("gettimeofday failed");
 }
 
 static void
@@ -300,6 +284,27 @@ get_open_fd_max(void)
 #endif
 }
 
+#ifndef HAVE_SETSID
+static void
+detach_controlling_tty(void)
+{
+#ifdef HAVE_TIOCNOTTY
+	int tty_fd;
+
+	tty_fd = open("/dev/tty", O_RDWR);
+
+	/* The current process does not have a controlling tty. */
+	if (tty_fd < 0)
+		return;
+
+	if (ioctl(tty_fd, TIOCNOTTY, 0) != 0)
+		fatal("unable to detach controlling tty");
+
+	close(tty_fd);
+#endif
+}
+#endif
+
 static void
 daemonize(void)
 {
@@ -310,7 +315,7 @@ daemonize(void)
 
 	pid = fork();
 	if (pid < 0)
-		fatal("Unable to do first fork.\n");
+		fatal("unable to do first fork");
 	else if (pid) /* Parent. */
 		_exit(0);
 
@@ -319,11 +324,12 @@ daemonize(void)
 	setsid();
 #else
 	setpgid(0, 0);
+	detach_controlling_tty();
 #endif
 
 	pid = fork();
 	if (pid < 0)
-		fatal("Unable to do second fork.\n");
+		fatal("unable to do second fork");
 	else if (pid) /* Parent. */
 		_exit(0);
 
@@ -332,7 +338,7 @@ daemonize(void)
 }
 
 static void
-push(struct pid_list **list, pid_t pid)
+pid_list_push(struct pid_list **list, pid_t pid)
 {
 	struct pid_list *p;
 
@@ -343,7 +349,7 @@ push(struct pid_list **list, pid_t pid)
 }
 
 static void
-clear(struct pid_list **list)
+pid_list_free(struct pid_list **list)
 {
 	struct pid_list *here, *next;
 
@@ -356,7 +362,7 @@ clear(struct pid_list **list)
 }
 
 static void
-do_help(void)
+usage(void)
 {
 	printf(
 "Usage: start-stop-daemon [<option> ...] <command>\n"
@@ -421,7 +427,7 @@ do_version(void)
 	printf("Written by Marek Michalkiewicz, public domain.\n");
 }
 
-static void
+static void DPKG_ATTR_NORET
 badusage(const char *msg)
 {
 	if (msg)
@@ -475,16 +481,16 @@ parse_integer(const char *string, int *value_r)
 }
 
 static int
-parse_signal(const char *signal_str, int *signal_nr)
+parse_signal(const char *sig_str, int *sig_num)
 {
 	unsigned int i;
 
-	if (parse_integer(signal_str, signal_nr) == 0)
+	if (parse_integer(sig_str, sig_num) == 0)
 		return 0;
 
-	for (i = 0; i < sizeof_array(siglist); i++) {
-		if (strcmp(signal_str, siglist[i].name) == 0) {
-			*signal_nr = siglist[i].signal;
+	for (i = 0; i < array_count(siglist); i++) {
+		if (strcmp(sig_str, siglist[i].name) == 0) {
+			*sig_num = siglist[i].signal;
 			return 0;
 		}
 	}
@@ -595,7 +601,7 @@ set_proc_schedule(struct res_schedule *sched)
 	param.sched_priority = sched->priority;
 
 	if (sched_setscheduler(getpid(), sched->policy, &param) == -1)
-		fatal("Unable to set process scheduler");
+		fatal("unable to set process scheduler");
 #endif
 }
 
@@ -615,7 +621,7 @@ set_io_schedule(struct res_schedule *sched)
 
 	io_sched_mask = IOPRIO_PRIO_VALUE(sched->policy, sched->priority);
 	if (ioprio_set(IOPRIO_WHO_PROCESS, getpid(), io_sched_mask) == -1)
-		warning("Unable to alter IO priority to mask %i (%s)\n",
+		warning("unable to alter IO priority to mask %i (%s)\n",
 		        io_sched_mask, strerror(errno));
 #endif
 }
@@ -690,6 +696,9 @@ parse_schedule(const char *schedule_str)
 			}
 			count++;
 		}
+		if (repeatat == count)
+			badusage("invalid schedule: 'forever' appears last, "
+			         "nothing to repeat");
 		if (repeatat >= 0) {
 			schedule[count].type = sched_goto;
 			schedule[count].value = repeatat;
@@ -745,7 +754,7 @@ parse_options(int argc, char * const *argv)
 			break;
 		switch (c) {
 		case 'H':  /* --help */
-			do_help();
+			usage();
 			exit(0);
 		case 'K':  /* --stop */
 			stop = 1;
@@ -787,7 +796,7 @@ parse_options(int argc, char * const *argv)
 			execname = optarg;
 			break;
 		case 'c':  /* --chuid <username>|<uid> */
-			/* we copy the string just in case we need the
+			/* We copy the string just in case we need the
 			 * argument later. */
 			changeuser = xstrdup(optarg);
 			changeuser = strtok(changeuser, ":");
@@ -1049,7 +1058,7 @@ pid_is_cmd(pid_t pid, const char *name)
 		fclose(f);
 		return false;
 	}
-	/* This hopefully handles command names containing ')'. */
+	/* This hopefully handles command names containing ‘)’. */
 	while ((c = getc(f)) != EOF && c == *name)
 		name++;
 	fclose(f);
@@ -1095,8 +1104,8 @@ pid_is_cmd(pid_t pid, const char *name)
 	if (pid_argv_p == NULL)
 		errx(1, "%s", kvm_geterr(kd));
 
-	start_argv_0_p = *pid_argv_p;
 	/* Find and compare string. */
+	start_argv_0_p = *pid_argv_p;
 
 	/* Find end of argv[0] then copy and cut of str there. */
 	end_argv_0_p = strchr(*pid_argv_p, ' ');
@@ -1134,12 +1143,12 @@ pid_is_running(pid_t pid)
 	else if (errno == ESRCH)
 		return false;
 	else
-		fatal("Error checking pid %u status: %s", pid, strerror(errno));
+		fatal("error checking pid %u status", pid);
 }
 #endif
 
 static void
-check(pid_t pid)
+pid_check(pid_t pid)
 {
 #if defined(OSLinux) || defined(OShpux)
 	if (execname && !pid_is_exec(pid, &exec_stat))
@@ -1158,7 +1167,7 @@ check(pid_t pid)
 		return;
 	if (start && !pid_is_running(pid))
 		return;
-	push(&found, pid);
+	pid_list_push(&found, pid);
 }
 
 static void
@@ -1168,17 +1177,17 @@ do_pidfile(const char *name)
 	static pid_t pid = 0;
 
 	if (pid) {
-		check(pid);
+		pid_check(pid);
 		return;
 	}
 
 	f = fopen(name, "r");
 	if (f) {
 		if (fscanf(f, "%d", &pid) == 1)
-			check(pid);
+			pid_check(pid);
 		fclose(f);
 	} else if (errno != ENOENT)
-		fatal("open pidfile %s: %s", name, strerror(errno));
+		fatal("unable to open pidfile %s", name);
 }
 
 #if defined(OSLinux) || defined (OSsunos)
@@ -1192,14 +1201,14 @@ do_procinit(void)
 
 	procdir = opendir("/proc");
 	if (!procdir)
-		fatal("opendir /proc: %s", strerror(errno));
+		fatal("unable to opendir /proc");
 
 	foundany = 0;
 	while ((entry = readdir(procdir)) != NULL) {
 		if (sscanf(entry->d_name, "%d", &pid) != 1)
 			continue;
 		foundany++;
-		check(pid);
+		pid_check(pid);
 	}
 	closedir(procdir);
 	if (!foundany)
@@ -1209,7 +1218,7 @@ do_procinit(void)
 static int
 check_proc_stat(struct proc_stat *ps)
 {
-	check(ps->pid);
+	pid_check(ps->pid);
 	return 0;
 }
 
@@ -1231,7 +1240,7 @@ do_procinit(void)
 
 	while ((count = pstat_getproc(pst, sizeof(pst[0]), 10, idx)) > 0) {
 		for (i = 0; i < count; i++)
-			check(pst[i].pst_pid);
+			pid_check(pst[i].pst_pid);
 		idx = pst[count - 1].pst_idx + 1;
 	}
 }
@@ -1246,7 +1255,7 @@ do_procinit(void)
 static void
 do_findprocs(void)
 {
-	clear(&found);
+	pid_list_free(&found);
 
 	if (pidfile)
 		do_pidfile(pidfile);
@@ -1255,8 +1264,7 @@ do_findprocs(void)
 }
 
 static void
-do_stop(int signal_nr, int quietmode, int *n_killed, int *n_notkilled,
-        int retry_nr)
+do_stop(int sig_num, int quiet, int *n_killed, int *n_notkilled, int retry_nr)
 {
 	struct pid_list *p;
 
@@ -1268,19 +1276,19 @@ do_stop(int signal_nr, int quietmode, int *n_killed, int *n_notkilled,
 	if (!found)
 		return;
 
-	clear(&killed);
+	pid_list_free(&killed);
 
 	for (p = found; p; p = p->next) {
 		if (testmode) {
 			if (quietmode <= 0)
 				printf("Would send signal %d to %d.\n",
-				       signal_nr, p->pid);
+				       sig_num, p->pid);
 			(*n_killed)++;
-		} else if (kill(p->pid, signal_nr) == 0) {
-			push(&killed, p->pid);
+		} else if (kill(p->pid, sig_num) == 0) {
+			pid_list_push(&killed, p->pid);
 			(*n_killed)++;
 		} else {
-			if (signal_nr)
+			if (sig_num)
 				warning("failed to kill %d: %s\n",
 				        p->pid, strerror(errno));
 			(*n_notkilled)++;
@@ -1304,11 +1312,89 @@ set_what_stop(const char *str)
 	what_stop[sizeof(what_stop) - 1] = '\0';
 }
 
+/*
+ * We want to keep polling for the processes, to see if they've exited, or
+ * until the timeout expires.
+ *
+ * This is a somewhat complicated algorithm to try to ensure that we notice
+ * reasonably quickly when all the processes have exited, but don't spend
+ * too much CPU time polling. In particular, on a fast machine with
+ * quick-exiting daemons we don't want to delay system shutdown too much,
+ * whereas on a slow one, or where processes are taking some time to exit,
+ * we want to increase the polling interval.
+ *
+ * The algorithm is as follows: we measure the elapsed time it takes to do
+ * one poll(), and wait a multiple of this time for the next poll. However,
+ * if that would put us past the end of the timeout period we wait only as
+ * long as the timeout period, but in any case we always wait at least
+ * MIN_POLL_INTERVAL (20ms). The multiple (‘ratio’) starts out as 2, and
+ * increases by 1 for each poll to a maximum of 10; so we use up to between
+ * 30% and 10% of the machine's resources (assuming a few reasonable things
+ * about system performance).
+ */
+static bool
+do_stop_timeout(int timeout, int *n_killed, int *n_notkilled)
+{
+	struct timeval stopat, before, after, interval, maxinterval;
+	int r, ratio;
+
+	xgettimeofday(&stopat);
+	stopat.tv_sec += timeout;
+	ratio = 1;
+	for (;;) {
+		xgettimeofday(&before);
+		if (timercmp(&before, &stopat, >))
+			return false;
+
+		do_stop(0, 1, n_killed, n_notkilled, 0);
+		if (!*n_killed)
+			return true;
+
+		xgettimeofday(&after);
+
+		if (!timercmp(&after, &stopat, <))
+			return false;
+
+		if (ratio < 10)
+			ratio++;
+
+		timersub(&stopat, &after, &maxinterval);
+		timersub(&after, &before, &interval);
+		tmul(&interval, ratio);
+
+		if (interval.tv_sec < 0 || interval.tv_usec < 0)
+			interval.tv_sec = interval.tv_usec = 0;
+
+		if (timercmp(&interval, &maxinterval, >))
+			interval = maxinterval;
+
+		if (interval.tv_sec == 0 &&
+		    interval.tv_usec <= MIN_POLL_INTERVAL)
+			interval.tv_usec = MIN_POLL_INTERVAL;
+
+		r = select(0, NULL, NULL, NULL, &interval);
+		if (r < 0 && errno != EINTR)
+			fatal("select() failed for pause");
+	}
+}
+
+static int
+finish_stop_schedule(bool anykilled)
+{
+	if (anykilled)
+		return 0;
+
+	if (quietmode <= 0)
+		printf("No %s found running; none killed.\n", what_stop);
+
+	return exitnodo;
+}
+
 static int
 run_stop_schedule(void)
 {
-	int r, position, n_killed, n_notkilled, value, ratio, anykilled, retry_nr;
-	struct timeval stopat, before, after, interval, maxinterval;
+	int position, n_killed, n_notkilled, value, retry_nr;
+	bool anykilled;
 
 	if (testmode) {
 		if (schedule != NULL) {
@@ -1327,9 +1413,9 @@ run_stop_schedule(void)
 	else if (userspec)
 		sprintf(what_stop, "process(es) owned by '%.200s'", userspec);
 	else
-		fatal("internal error, please report");
+		fatal("internal error, no match option, please report");
 
-	anykilled = 0;
+	anykilled = false;
 	retry_nr = 0;
 
 	if (schedule == NULL) {
@@ -1337,92 +1423,34 @@ run_stop_schedule(void)
 		if (n_notkilled > 0 && quietmode <= 0)
 			printf("%d pids were not killed\n", n_notkilled);
 		if (n_killed)
-			anykilled = 1;
-		goto x_finished;
+			anykilled = true;
+		return finish_stop_schedule(anykilled);
 	}
 
-	for (position = 0; position < schedule_length; ) {
+	for (position = 0; position < schedule_length; position++) {
+	reposition:
 		value = schedule[position].value;
 		n_notkilled = 0;
 
 		switch (schedule[position].type) {
 		case sched_goto:
 			position = value;
-			continue;
+			goto reposition;
 		case sched_signal:
 			do_stop(value, quietmode, &n_killed, &n_notkilled, retry_nr++);
 			if (!n_killed)
-				goto x_finished;
+				return finish_stop_schedule(anykilled);
 			else
-				anykilled = 1;
-			goto next_item;
+				anykilled = true;
+			continue;
 		case sched_timeout:
- /* We want to keep polling for the processes, to see if they've exited,
-  * or until the timeout expires.
-  *
-  * This is a somewhat complicated algorithm to try to ensure that we
-  * notice reasonably quickly when all the processes have exited, but
-  * don't spend too much CPU time polling.  In particular, on a fast
-  * machine with quick-exiting daemons we don't want to delay system
-  * shutdown too much, whereas on a slow one, or where processes are
-  * taking some time to exit, we want to increase the polling
-  * interval.
-  *
-  * The algorithm is as follows: we measure the elapsed time it takes
-  * to do one poll(), and wait a multiple of this time for the next
-  * poll.  However, if that would put us past the end of the timeout
-  * period we wait only as long as the timeout period, but in any case
-  * we always wait at least MIN_POLL_INTERVAL (20ms).  The multiple
-  * (‘ratio’) starts out as 2, and increases by 1 for each poll to a
-  * maximum of 10; so we use up to between 30% and 10% of the
-  * machine's resources (assuming a few reasonable things about system
-  * performance).
-  */
-			xgettimeofday(&stopat);
-			stopat.tv_sec += value;
-			ratio = 1;
-			for (;;) {
-				xgettimeofday(&before);
-				if (timercmp(&before, &stopat, >))
-					goto next_item;
-
-				do_stop(0, 1, &n_killed, &n_notkilled, 0);
-				if (!n_killed)
-					goto x_finished;
-
-				xgettimeofday(&after);
-
-				if (!timercmp(&after, &stopat, <))
-					goto next_item;
-
-				if (ratio < 10)
-					ratio++;
-
-				timersub(&stopat, &after, &maxinterval);
-				timersub(&after, &before, &interval);
-				tmul(&interval, ratio);
-
-				if (interval.tv_sec < 0 || interval.tv_usec < 0)
-					interval.tv_sec = interval.tv_usec = 0;
-
-				if (timercmp(&interval, &maxinterval, >))
-					interval = maxinterval;
-
-				if (interval.tv_sec == 0 &&
-				    interval.tv_usec <= MIN_POLL_INTERVAL)
-					interval.tv_usec = MIN_POLL_INTERVAL;
-
-				r = select(0, NULL, NULL, NULL, &interval);
-				if (r < 0 && errno != EINTR)
-					fatal("select() failed for pause: %s",
-					      strerror(errno));
-			}
+			if (do_stop_timeout(value, &n_killed, &n_notkilled))
+				return finish_stop_schedule(anykilled);
+			else
+				continue;
 		default:
 			assert(!"schedule[].type value must be valid");
 		}
-
-	next_item:
-		position++;
 	}
 
 	if (quietmode <= 0)
@@ -1430,15 +1458,6 @@ run_stop_schedule(void)
 		       what_stop, n_killed);
 
 	return 2;
-
-x_finished:
-	if (!anykilled) {
-		if (quietmode <= 0)
-			printf("No %s found running; none killed.\n", what_stop);
-		return exitnodo;
-	} else {
-		return 0;
-	}
 }
 
 int
@@ -1447,9 +1466,6 @@ main(int argc, char **argv)
 	int devnull_fd = -1;
 	gid_t rgid;
 	uid_t ruid;
-#ifdef HAVE_TIOCNOTTY
-	int tty_fd = -1;
-#endif
 	progname = argv[0];
 
 	parse_options(argc, argv);
@@ -1470,7 +1486,7 @@ main(int argc, char **argv)
 			fullexecname = execname;
 
 		if (stat(fullexecname, &exec_stat))
-			fatal("stat %s: %s", fullexecname, strerror(errno));
+			fatal("unable to stat %s", fullexecname);
 
 		if (fullexecname != execname)
 			free(fullexecname);
@@ -1481,7 +1497,7 @@ main(int argc, char **argv)
 
 		pw = getpwnam(userspec);
 		if (!pw)
-			fatal("user '%s' not found\n", userspec);
+			fatal("user '%s' not found", userspec);
 
 		user_id = pw->pw_uid;
 	}
@@ -1489,25 +1505,26 @@ main(int argc, char **argv)
 	if (changegroup && sscanf(changegroup, "%d", &runas_gid) != 1) {
 		struct group *gr = getgrnam(changegroup);
 		if (!gr)
-			fatal("group '%s' not found\n", changegroup);
+			fatal("group '%s' not found", changegroup);
 		runas_gid = gr->gr_gid;
 	}
 	if (changeuser) {
 		struct passwd *pw;
+		struct stat st;
 
 		if (sscanf(changeuser, "%d", &runas_uid) == 1)
 			pw = getpwuid(runas_uid);
 		else
 			pw = getpwnam(changeuser);
 		if (!pw)
-			fatal("user '%s' not found\n", changeuser);
+			fatal("user '%s' not found", changeuser);
 		runas_uid = pw->pw_uid;
 		if (changegroup == NULL) {
 			/* Pass the default group of this user. */
 			changegroup = ""; /* Just empty. */
 			runas_gid = pw->pw_gid;
 		}
-		if (access(pw->pw_dir, F_OK) == 0)
+		if (stat(pw->pw_dir, &st) == 0)
 			setenv("HOME", pw->pw_dir, 1);
 	}
 
@@ -1555,16 +1572,14 @@ main(int argc, char **argv)
 		/* Ok, we need to detach this process. */
 		daemonize();
 
-#ifdef HAVE_TIOCNOTTY
-		tty_fd = open("/dev/tty", O_RDWR);
-#endif
 		devnull_fd = open("/dev/null", O_RDWR);
+		if (devnull_fd < 0)
+			fatal("unable to open '%s'", "/dev/null");
 	}
 	if (nicelevel) {
 		errno = 0;
 		if ((nice(nicelevel) == -1) && (errno != 0))
-			fatal("Unable to alter nice level by %i: %s",
-			      nicelevel, strerror(errno));
+			fatal("unable to alter nice level by %i", nicelevel);
 	}
 	if (proc_sched)
 		set_proc_schedule(proc_sched);
@@ -1573,34 +1588,31 @@ main(int argc, char **argv)
 	if (umask_value >= 0)
 		umask(umask_value);
 	if (mpidfile && pidfile != NULL) {
-		/* User wants _us_ to make the pidfile. :) */
+		/* User wants _us_ to make the pidfile. */
 		FILE *pidf = fopen(pidfile, "w");
 		pid_t pidt = getpid();
 		if (pidf == NULL)
-			fatal("Unable to open pidfile '%s' for writing: %s",
-			      pidfile, strerror(errno));
+			fatal("unable to open pidfile '%s' for writing",
+			      pidfile);
 		fprintf(pidf, "%d\n", pidt);
-		if (fflush(pidf))
-			fatal("unable to flush pidfile '%s'", pidfile);
-		if (fsync(fileno(pidf)))
-			fatal("unable to sync pidfile '%s'", pidfile);
-		fclose(pidf);
+		if (fclose(pidf))
+			fatal("unable to close pidfile '%s'", pidfile);
 	}
 	if (changeroot != NULL) {
 		if (chdir(changeroot) < 0)
-			fatal("Unable to chdir() to %s", changeroot);
+			fatal("unable to chdir() to %s", changeroot);
 		if (chroot(changeroot) < 0)
-			fatal("Unable to chroot() to %s", changeroot);
+			fatal("unable to chroot() to %s", changeroot);
 	}
 	if (chdir(changedir) < 0)
-		fatal("Unable to chdir() to %s", changedir);
+		fatal("unable to chdir() to %s", changedir);
 
 	rgid = getgid();
 	ruid = getuid();
 	if (changegroup != NULL) {
 		if (rgid != (gid_t)runas_gid)
 			if (setgid(runas_gid))
-				fatal("Unable to set gid to %d", runas_gid);
+				fatal("unable to set gid to %d", runas_gid);
 	}
 	if (changeuser != NULL) {
 		/* We assume that if our real user and group are the same as
@@ -1608,24 +1620,21 @@ main(int argc, char **argv)
 		 * will be already in place. */
 		if (rgid != (gid_t)runas_gid || ruid != (uid_t)runas_uid)
 			if (initgroups(changeuser, runas_gid))
-				fatal("Unable to set initgroups() with gid %d",
+				fatal("unable to set initgroups() with gid %d",
 				      runas_gid);
 
 		if (ruid != (uid_t)runas_uid)
 			if (setuid(runas_uid))
-				fatal("Unable to set uid to %s", changeuser);
+				fatal("unable to set uid to %s", changeuser);
 	}
 
 	if (background) {
-		/* Continue background setup. */
 		int i;
-#ifdef HAVE_TIOCNOTTY
-		 /* Change tty. */
-		ioctl(tty_fd, TIOCNOTTY, 0);
-		close(tty_fd);
-#endif
+
+		/* Set a default umask for dumb programs. */
 		if (umask_value < 0)
-			umask(022); /* Set a default for dumb programs. */
+			umask(022);
+
 		dup2(devnull_fd, 0); /* stdin */
 		dup2(devnull_fd, 1); /* stdout */
 		dup2(devnull_fd, 2); /* stderr */
@@ -1635,6 +1644,5 @@ main(int argc, char **argv)
 			close(i);
 	}
 	execv(startas, argv);
-	fatal("Unable to start %s: %s", startas, strerror(errno));
+	fatal("unable to start %s", startas);
 }
-
